﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using Kaleida.ServiceMonitor.Model.Layout;
using NUnit.Framework;

namespace Kaleida.UnitTests.Model.Layout
{
    public class StatusWindowLayoutStrategyTests
    {
        [Test]
        public void TestSingleScriptUsesFullArea()
        {
            var rootNode = new ScriptNode("System1");
            var strategy = new NestedTilesLayoutStrategy();

            var scriptAreas = strategy.CalculateLayout(rootNode, new Rectangle(10, 20, 810, 620));
            
            Assert.AreEqual(1, scriptAreas.Count);
            Assert.AreEqual("System1: Top-left=(10,20) Rectangle=(810x620)", scriptAreas[0].ToString());
        }

        [Test]
        public void TestTwoScriptsAtTopLevelUsesAreaSplitHorizontally()
        {
            var rootNode = new ScriptNode("Root");
            rootNode.AddChild(new ScriptNode("System1"));
            rootNode.AddChild(new ScriptNode("System2"));

            var strategy = new NestedTilesLayoutStrategy();

            var scriptAreas = strategy.CalculateLayout(rootNode, new Rectangle(10, 20, 810, 620));
            
            Assert.AreEqual(2, scriptAreas.Count);
            Assert.AreEqual("System1: Top-left=(10,20) Rectangle=(810x310)", scriptAreas[0].ToString());
            Assert.AreEqual("System2: Top-left=(10,330) Rectangle=(810x310)", scriptAreas[1].ToString());
        }

        [Test]
        public void TestTwoScriptsAtTopLevelUseAreaSplitVertically()
        {
            var rootNode = new ScriptNode("Root");
            var node1 = new ScriptNode("Group");
            rootNode.AddChild(node1);
            node1.AddChild(new ScriptNode("System1"));
            node1.AddChild(new ScriptNode("System2"));

            var strategy = new NestedTilesLayoutStrategy();

            var scriptAreas = strategy.CalculateLayout(rootNode, new Rectangle(10, 20, 810, 620));

            Assert.AreEqual(2, scriptAreas.Count);
            Assert.AreEqual("System1: Top-left=(10,20) Rectangle=(405x620)", scriptAreas[0].ToString());
            Assert.AreEqual("System2: Top-left=(415,20) Rectangle=(405x620)", scriptAreas[1].ToString());
        }

        [Test]
        public void TestRemainderIsAllocatedToFinalSplit()
        {
            var rootNode = new ScriptNode("Root");
            rootNode.AddChild(new ScriptNode("System1"));
            rootNode.AddChild(new ScriptNode("System2"));
            rootNode.AddChild(new ScriptNode("System3"));

            var strategy = new NestedTilesLayoutStrategy();

            var scriptAreas = strategy.CalculateLayout(rootNode, new Rectangle(0, 0, 150, 100));

            Assert.AreEqual(3, scriptAreas.Count);
            Assert.AreEqual("System1: Top-left=(0,0) Rectangle=(150x33)", scriptAreas[0].ToString());
            Assert.AreEqual("System2: Top-left=(0,33) Rectangle=(150x33)", scriptAreas[1].ToString());
            Assert.AreEqual("System3: Top-left=(0,66) Rectangle=(150x34)", scriptAreas[2].ToString());
        }

        [Test]
        public void TestNestedScriptsAlternateBetweenHorizontalAndVerticalSubDivision()
        {
            var rootNode = new ScriptNode("Root");
            
            var system1 = new ScriptNode("System1");
            var system1A = new ScriptNode("System1a");
            system1A.AddChild(new ScriptNode("System1a - i"));
            system1A.AddChild(new ScriptNode("System1a - ii"));
            system1.AddChild(system1A);
            system1.AddChild(new ScriptNode("System1b"));
            system1.AddChild(new ScriptNode("System1c"));
            rootNode.AddChild(system1);
            
            var system2 = new ScriptNode("System2");
            rootNode.AddChild(system2);

            var strategy = new NestedTilesLayoutStrategy();

            var scriptAreas = strategy.CalculateLayout(rootNode, new Rectangle(10, 20, 810, 620));

            Trace.WriteLine(string.Join("\r\n", scriptAreas));

            Assert.AreEqual(5, scriptAreas.Count);
            Assert.AreEqual("System1a - i: Top-left=(10,20) Rectangle=(270x155)", scriptAreas[0].ToString());
            Assert.AreEqual("System1a - ii: Top-left=(10,175) Rectangle=(270x155)", scriptAreas[1].ToString());
            Assert.AreEqual("System1b: Top-left=(280,20) Rectangle=(270x310)", scriptAreas[2].ToString());
            Assert.AreEqual("System1c: Top-left=(550,20) Rectangle=(270x310)", scriptAreas[3].ToString());
            Assert.AreEqual("System2: Top-left=(10,330) Rectangle=(810x310)", scriptAreas[4].ToString());
        }

        private class ScriptNode : IScriptNode
        {
            private readonly string name;
            private readonly List<IScriptNode> childNodes = new List<IScriptNode>();

            public ScriptNode(string name)
            {
                this.name = name;
            }

            public void AddChild(IScriptNode node)
            {
                childNodes.Add(node);
            }

            public IList<IScriptNode> Children
            {
                get { return childNodes; }
            }

            public override string ToString()
            {
                return name;
            }
        }
    }
}
